//gcc roto.c -o roto -lSDL -lm

#include <stdio.h>
#include <stdlib.h>
#include <SDL/SDL.h>
#include <math.h>

#define SCREEN_WIDTH 320
#define SCREEN_HEIGHT 240

#define PI 3.1415927

typedef uint8_t u8;
typedef uint16_t u16;
typedef uint32_t u32;

typedef int8_t s8;
typedef int16_t s16;
typedef int32_t s32;

//Global variables
int quit=0;
SDL_Event event;
SDL_Surface *screen;

SDL_Surface *pRotoPic;			// Ptr sur l'image pour rotozoom.

s16	pSinCos[256 + 64];	// Table contenant les sin et cos * 256, sur 256 angles.
s16	*pSin;			// Ptrs sur les tables.
s16	*pCos;

SDL_Color	pColors[256];	// Palette générale, à réinitialiser au changement de mode.

void Rotator(void);
void RotatorInit(void);

// Variables à la con.
struct SRoto
{
	u8	nAngle;		// Angle.
	u32	nAngVar;	// 1 = Angle++, 3 = Angle--, 0 et 2 = scroll.
	u32	nAngCnt;	// Compteur.

	u8	nZoom;		// En fait index pour taper dans la table de sinus.
	u32	nZoomVar;	// 0 = Zoome, 1 = Stop.
	u32	nZoomCnt;	// Compteur.

	u32	nScroll;	// Offset du scroll dans la planche.

};
struct SRoto	gRoto;

// Initialisation.
void RotatorInit(void)
{
	gRoto.nAngle = 192;
	gRoto.nAngVar = rand() & 3;
	gRoto.nAngCnt = (rand() & 63) | 16;

	gRoto.nZoom = 64;
	gRoto.nZoomVar = rand() & 1;
	gRoto.nZoomCnt = (rand() & 63) | 32;

	gRoto.nScroll = 0;
}

// The rotator.
#define	RZ_TYPE	1	// 0 = RotoZoom classique / 1 = Avec effet saccadé.

#if RZ_TYPE	== 0
    #define	ROTO_CTR_X	(SCREEN_WIDTH / 2)		// Centrage au milieu de l'écran.
    #define	ROTO_CTR_Y	(SCREEN_HEIGHT/ 2)
#else
    #define	ROTO_CTR_X	231					// Centrage sur le texte du menu.
    #define	ROTO_CTR_Y	158
#endif

void Rotator(void)
{
	u32	ix, iy;
	u16	px, py, px2, py2, incx, incy;
	u32	nMultiplier;
	u8	*pScr, *pPic;

	SDL_LockSurface(screen);

	// Multiplicateur (Zoom).
    #if RZ_TYPE	== 0
        nMultiplier = 0x08000 + ((0x0060 + pCos[(gRoto.nAngle + 192) & 0xFF]/2) << 9);
    #else
        nMultiplier = 0x08000 + ((0x0060 + pCos[gRoto.nZoom]/2) << 9);
    #endif

	// Val sin et cos (Roto).
	incx = (pCos[gRoto.nAngle] * nMultiplier) >> 16;
	incy = (pSin[gRoto.nAngle] * nMultiplier) >> 16;

	// Point de départ.
	// x' = x cos f - y sin f
	// y' = y cos f + x sin f
	px = (128 << 8) - ((ROTO_CTR_X * incx) - (ROTO_CTR_Y * incy));
	py = (128 << 8) - ((ROTO_CTR_Y * incx) + (ROTO_CTR_X * incy));

	// Loop.
	pScr = (u8*)screen->pixels;
	pPic = (u8*)pRotoPic->pixels;
	pPic += (gRoto.nScroll & 63) << 8;	// Scroll.
	s32	nPitchDiff = screen->pitch - SCREEN_WIDTH;
    
	for (iy = 0; iy < SCREEN_HEIGHT; iy++)
	{
		px2 = px;
		py2 = py;

		for (ix = 0; ix < SCREEN_WIDTH; ix++)
		{
			*pScr++ = *(pPic + ((py2 >> 8) << 8) + (px2 >> 8));

			px2 += incx;
			py2 += incy;
		}
		pScr += nPitchDiff;

		//incx++;		// Déformation.

		px -= incy;
		py += incx;
	}

    #if RZ_TYPE	== 0
        gRoto.nAngle += 1;
    #else
        // Déplacement des rotations, zooms, scrolls...
        gRoto.nAngCnt--;
        gRoto.nAngCnt &= 63;
        if (gRoto.nAngCnt == 0)
        {
            gRoto.nAngVar = (gRoto.nAngVar + 1) & 3;
            gRoto.nAngCnt = (rand() & 63) | 16;
        }
        if (gRoto.nAngVar == 1) gRoto.nAngle++;
        else if (gRoto.nAngVar == 3) gRoto.nAngle--;
        else gRoto.nScroll++;

        gRoto.nZoomCnt--;
        gRoto.nZoomCnt &= 63;
        if (gRoto.nZoomCnt == 0)
        {
            gRoto.nZoomVar = (gRoto.nZoomVar + 1) & 1;
            gRoto.nZoomCnt = (rand() & 63) | 32;
        }
        if (gRoto.nZoomVar == 0) gRoto.nZoom++;
    #endif

	SDL_UnlockSurface(screen);

}

// Précalcul des tables de sinus-cosinus.
// 256 angles, val *256 (=> varie de -256 à 256).
void PrecaSinCos(void)
{
	u32	i;

	for (i = 0; i < 256 + 64; i++)
	{
		pSinCos[i] = (s16) (cos(i * 2 * PI / 256) * 256);
		//printf("i = %d : %f\n", i, cos(i * 2 * PI / 256) * 256);
		//printf("i = %d : %d\n", i, gVar.pSinCos[i]);
	}
	pSin = pSinCos + 64;
	pCos = pSinCos;

	/*
	for (i = 0; i < 256; i++)
	{
		printf("i = %d : sin = %d - cos = %d\n", i, gVar.pSin[i], gVar.pCos[i]);
	}
	*/

}

// Création de la palette :
// On recopie dans la palette générale la partie de palette correspondant au décor + la palette des sprites.
void SetPalette(SDL_Surface *pBkg, SDL_Color *pSprPal, u32 nSprPalIdx)
{
	u32	i;

	SDL_Color	*pSrcPal = pBkg->format->palette->colors;

	// Couleurs du décor.
	for (i = 0; i < nSprPalIdx; i++)
	{
		pColors[i] = pSrcPal[i];
	}
        
	//~ // Couleurs des sprites.
	//~ for (i=0; i < 256; i++)
	//~ {
		//~ pColors[i] = pSprPal[i - 32];
        
	//~ }
	// Palette logique.
	SDL_SetPalette(screen, SDL_LOGPAL, pColors, 0, 256);

}

int mainLoop()
{
    //main game loop
    while(!quit)
    {
        while(SDL_PollEvent(&event))
        {
            if (event.type == SDL_QUIT ||(event.type == SDL_KEYDOWN && event.key.keysym.sym == SDLK_ESCAPE))
            quit = 1;        /* Window closed */
        }
        
        Rotator();
        //update the screen
        SDL_Flip(screen);
        SDL_Delay(10);
    }

    return 0;
}

int main()
{
    // init video stuff
    // init screen
    screen = SDL_SetVideoMode(SCREEN_WIDTH, SCREEN_HEIGHT, 8, SDL_SWSURFACE);
    if(screen == NULL)
    {
        fprintf(stderr, "Can't initialize SDL: %s\n", SDL_GetError());
        exit(-1);
    }
    
    SDL_WM_SetCaption("Roto", "Roto");

    // Image pour le rotozoom.
    if ((pRotoPic = SDL_LoadBMP("gfx/title_roto.bmp")) == NULL) {
        fprintf(stderr, "Couldn't load picture '%s' : %s\n", "gfx/title_roto.bmp", SDL_GetError());
        exit(1);
    }

    SetPalette(pRotoPic, pColors, 256);

    PrecaSinCos();
    RotatorInit();
 
    //main game loop
    mainLoop();

    SDL_FreeSurface(pRotoPic);
    
    SDL_Quit();
    return 0;
}
